"use strict";

module.exports = function(babel) {
    const t = babel.types;

    return new babel.Plugin("babel-ui5-plugin", {
        visitor:
        {
            Program : {
                exit(node, parent, scope, state)
                {
                    this.state.moduleFormatter.hasLocalExports = false;
                }
            },



            ImportDeclaration(node, parent, scope, state)
            {
                let path = node.source.value;
                let name = null;
                if (!node.specifier)
                {
                    const parts = path.split("/");
                    name = parts[parts.length - 1];
                }
                else
                {
                    name = node.specifiers[0].local.name;
                }

                if (!state.ui5)
                {
                    state.ui5 = { imports: [] };
                }
                state.ui5.imports.push({
                    name: name,
                    path: path
                });

                if (node.leadingComments)
                {
                    state.ui5.leadingComments = node.leadingComments;
                }

                this.dangerouslyRemove();
            },






            ExportDeclaration(node, parent, scope, state)
            {
                const depPaths = [];
                const depClasses = [];
                if (state.ui5 && state.ui5.imports)
                {
                    state.ui5.imports.forEach((imp) => {
                        depPaths.push(imp.path);
                        depClasses.push(imp.name);
                    });
                }

                const defineCallArgs = [
                    t.arrayExpression(depPaths.map(t.literal)),
                    t.functionExpression(null, depClasses.map(t.identifier), t.blockStatement([
                        t.expressionStatement(t.literal("use strict")),
                        t.returnStatement(_transformClassDeclaration(node.declaration, parent, scope, state, t))
                    ]))
                ];

                const defineCall = t.callExpression(t.identifier("\n\nsap.ui.define"), defineCallArgs);

                let leadingComments = [];
                if (state.ui5 && state.ui5.leadingComments)
                {
                    leadingComments = state.ui5.leadingComments;
                    if (node.leadingComments)
                    {
                        node.leadingComments.forEach(comment => {
                            leadingComments.push(comment);
                        });
                    }
                }

                const globals = [ "sap", "Symbol" ];
                if (state.ui5.namespace)
                {
                    var rootNamespace = state.ui5.namespace;
                    if (rootNamespace.indexOf(".") !== -1)
                    {
                        const pos = rootNamespace.indexOf(".");
                        globals.push(rootNamespace.substr(0, pos));
                    }
                }
                if (scope.globals)
                {
                    for (let key in scope.globals)
                    {
                        if (["arguments", "parseInt", "parseFloat", "undefined", "console"].indexOf(key) === -1 && globals.indexOf(key) === -1)
                        {
                            globals.push(key);
                        }
                    }
                }

                [
                    "eslint eol-last: 0",
                    "eslint no-underscore-dangle: 0",
                    "eslint no-unused-vars: 0",
                    "eslint no-undef-init: 0",
                    "eslint camelcase: 0",
                    "global " + globals.join(", ")
                ].forEach(comment => {
                    leadingComments.push({
                        type: "CommentBlock",
                        value: " " + comment + " "
                    });
                });

                node.leadingComments = leadingComments;

                return defineCall;
            },



            MemberExpression(node, parent, scope, state)
            {
                if (state.ui5 && state.ui5.fullClassName)
                {
                    if (node.object.name === state.ui5.className)
                    {
                        node.object.name = state.ui5.fullClassName;
                    }
                }
            },



            Super(node, parent, scope, state)
            {
                if (!state.ui5) return;

                if (!state.ui5.className || !state.ui5.superClassName)
                {
                    this.errorWithNode("The keyword 'super' can only used in a derrived class.");
                }

                return t.identifier("super");
            },


            ClassProperty(node, parent, scope, state)
            {
                if (node.typeAnnotation)
                {
                    const result = node.typeAnnotation.typeAnnotation;
                    if (result)
                    {
                        _transformTypeAnnotation(result);
                        return result;
                    }
                }
                else
                {
                    return t.literal(null);
                }
            },


            CallExpression(node, parent, scope, state)
            {
                if (!state.ui5) return;
                if (!node.callee) return;

                if (node.callee.type === "Super")
                {
                    if (!state.ui5.superClassName)
                    {
                        this.errorWithNode("The keyword 'super' can only used in a derrived class.");
                    }

                    const identifier = t.identifier(state.ui5.superClassName + ".apply");
                    let args = t.arrayExpression(node.arguments);
                    if (node.arguments.length === 1 && node.arguments[0].type === "Identifier" && node.arguments[0].name === "arguments")
                    {
                        args = t.identifier("arguments");
                    }
                    return t.callExpression(identifier, [
                        t.identifier("this"),
                        args
                    ]);
                }
                else if (node.callee.object && node.callee.object.type === "Super")
                {
                    if (!state.ui5.superClassName)
                    {
                        this.errorWithNode("The keyword 'super' can only used in a derrived class.");
                    }

                    const identifier = t.identifier(state.ui5.superClassName + ".prototype" + "." + node.callee.property.name + ".apply");
                    return t.callExpression(identifier, [
                        t.identifier("this"),
                        t.arrayExpression(node.arguments)
                    ]);
                }
            }
        }
    });
};




function _transformClassDeclaration(node, parent, scope, state, t)
{
    if (node.type !== "ClassDeclaration" || !state.ui5)
    {
        return node;
    }

    const className = node.id.name;
    const superClassName = node.superClass.name;
    if (state.ui5 === null)
    {
        state.ui5 = {};
    }
    state.ui5.className = className;
    state.ui5.superClassName = superClassName;

    if (state.opts.sourceFileName.indexOf("src/") === 0)
    {
        const filePath = state.opts.sourceFileName.substr(4);
        state.ui5.namespace = filePath.substr(0, filePath.lastIndexOf("/")).replace(new RegExp("/", "g"), ".");
        state.ui5.fullClassName = state.ui5.namespace + "." + state.ui5.className;
    }
    else if (state.opts.sourceFileName.indexOf("./src/") !== -1)
    {
        const filePath = state.opts.sourceFileName.substr(state.opts.sourceFileName.indexOf("./src/") + 6);
        state.ui5.namespace = filePath.substr(0, filePath.lastIndexOf("/")).replace(new RegExp("/", "g"), ".");
        state.ui5.fullClassName = state.ui5.namespace + "." + state.ui5.className;
    }
    else
    {
        state.ui5.namespace = null;
        state.ui5.fullClassName = state.ui5.className;
    }

    const props = [];
    node.body.body.forEach(member => {
        if (member.type === "MethodDefinition")
        {
            if (!member.static)
            {
                props.push(t.property("init", member.key, member.value));
            }
            else
            {
                const method = member.value;
                method.body.body.unshift(t.expressionStatement(t.literal("use strict")));

                const id = member.key;
                id.name = state.ui5.fullClassName + "." + id.name;
                const body = state.ast.program.body;
                body.push(t.expressionStatement(t.assignmentExpression("=", id, method)));
            }
        }
        else if (member.type === "ClassProperty")
        {
            if (!member.static)
            {
                props.push(t.property("init", member.key, member));
            }
            else
            {
                const prop = member;
                const id = prop.key;
                id.name = state.ui5.fullClassName + "." + id.name;
                let value = t.literal(null);
                if (prop.typeAnnotation && prop.typeAnnotation.typeAnnotation)
                {
                    value = prop.typeAnnotation.typeAnnotation
                    _transformTypeAnnotation(value);
                }
                const body = state.ast.program.body;
                body.push(t.expressionStatement(t.assignmentExpression("=", id, value)));
            }
        }
    });

    const bodyJSON = t.objectExpression(props);
    const extendCallArgs = [
        t.literal(state.ui5.fullClassName),
        bodyJSON
    ];
    const extendCall = t.callExpression(t.identifier(superClassName + ".extend"), extendCallArgs);
    return extendCall;
}

function _transformTypeAnnotation(typeAnnotation)
{
    switch (typeAnnotation.type)
    {
        case "ObjectTypeAnnotation":
            typeAnnotation.type = "ObjectExpression";
            typeAnnotation.properties.forEach(prop => {
                prop.type = "Property";
                _transformTypeAnnotation(prop.value);
            });
            break;
        case "TupleTypeAnnotation":
            typeAnnotation.type = "ArrayExpression";
            typeAnnotation.elements = typeAnnotation.types;
            typeAnnotation.elements.forEach(element => {
                _transformTypeAnnotation(element);
            });
            break;
        case "StringLiteralTypeAnnotation":
        case "NumberLiteralTypeAnnotation":
        case "BooleanLiteralTypeAnnotation":
            typeAnnotation.type = "Literal";
            break;
    }
}
